/*
  File: Activity.java

  Originally written by Keld Helsgaun and released into the public domain.
  This may be used for any purposes whatsoever without acknowledgment.

  History:
  Date       Who                What
  5May2000   kh         Created public version
*/

package simulation.activity;

/**
* This class may be used for activity-based discrete event 
* simulation.
* <p>
* An activity is associated with a start condition,
* a specification of its duration, and
* some start and finish actions. 
* <p>
* The start actions of an activity will be executed as soon as 
* its associated condition becomes true. 
* The finish actions will be executed when the activity ends 
* (after a time period equal to the duration of the activity).
* <p> 
* Activities are created as instances of <tt>Activity</tt>-derived
* classes that override the four abstract methods <tt>condition</tt>,
* <tt>startActions</tt>, <tt>duration</tt> and <tt>finishActions</tt>.
* 
* @see simulation.activity.Simulation
*/
public abstract class Activity {
    /**
    * The start condition for this activity.
    */
    protected abstract boolean condition();	
    
    
    /**
    * The start actions of this activity.
    */
    protected abstract void startActions();
    
    
    /**
    * The duration of this activity.
    */		
    protected abstract double duration();
    
    
    /**
    * The finish actions for this activity.
    */
    protected abstract void finishActions();

    /**
    * This constructor creates an <tt>Activity</tt>
    * and schedules it. 
    */	
    public Activity() { 
        schedule(); 
    } 
    
    /**
    * This constructor creates an <tt>Activity</tt>
    * but does not schedule it. It is used for creating the
    * heads of the event and wait lists. 
    */
    
    Activity(boolean dummy) {}
    
    /**
    * Returns the current simulation time.
    */	
    public final static double time() { return time; }
    
    /**
    * Schedules this activity.<br>
    * Its start time depends on its <tt>condition</tt>.
    * <p>
    * The activity is inserted at the back of the list of 
    * waiting activities.
    */  
    public final void schedule() {
        cancel(); 
        suc = waitList.suc;
        suc.pred = waitList.suc = this;
        pred = waitList;
    }
    
    /**
    * Causes this activity to <b>finish</b> at at the 
    * specified event time.
    * <p>
    * The activity is inserted into the event list at a
    * position corresponding to the current simulation time
    * and <b>after</b> any processes with the same event time.
    *
    * @param <tt>evTime</tt> The event time. 
    *
    * @exception <tt>RuntimeException</tt>
    * if <tt>evTime</tt> &lt; <tt>time()</tt> 
    */  
    private final void schedule(final double evTime) {
        if (evTime < time)
            throw new RuntimeException("attempt to schedule event in the past");
        cancel();
        eventTime = evTime;
        Activity a = SQS.pred; 
        while (a.eventTime > eventTime)
            a = a.pred;
        pred = a; 
        suc = a.suc;
        a.suc = suc.pred = this;
    }
    
    /**
    * Cancels this scheduled activity.
    * <p>
    * The activity is removed from the the list
    * of which it is a member (the event list or the wait list).
    * <p>
    * If it is not scheduled, the call has no effect.
    */	
    public final void cancel() {
        if (suc != null) {
            suc.pred = pred;
            pred.suc = suc;
            pred = suc = null;
        }
    }

    /**
    * Runs a simulation for a specified period of simulated time. 
    * <p>
    * Time will start at <tt>0</tt> and jump from event time to event time 
    * until either this period is over, there are no more actions 
    * to be executed, or the stopSimulation() method is called.
    * <p>
    * All activities waiting to <b>start</b> are held in a list, 
    * <tt>waitList</tt>.
    * All activities that are scheduled to <b>finish</b> are held 
    * in a list, <tt>SQS</tt>, ordered by their associated finish times. 
    * <p>
    * At each event time, the wait list is examined to see whether any
    * activity may start. 
    * In that case, the activity is removed from the list, its start 
    * actions are executed, and a finish event is scheduled to occur 
    * when the activity finishes.
    * <p> 
    * When no more activities may be started, time advances to the next
    * imminent event, and the associated finish actions are executed.
    * <p> 
    * This continues until the simulation ends.
    *  
    * @param <tt>period</tt> The length of the simulation period.
    */	
    public final static void runSimulation(final double period) {
        while (true) {
            for (Activity a = waitList.suc; 
                 a != waitList; 
                 a = a.suc) {
                if (a.condition()) {
                    a.cancel();
                    a.schedule(time + a.duration());
                    a.startActions();
                    a = waitList;
                }
            }
            if (SQS.suc == SQS)
                break;
            Activity a = SQS.suc;
            time = a.eventTime;
            a.cancel();
            if (time > period)
                break;
            a.finishActions();
        }
        stopSimulation();	
    }

    /**
    * Stops a simulation.
    */	
    public final static void stopSimulation() {
        while (waitList.suc != waitList) 
            waitList.suc.cancel();
        while (SQS.suc != SQS) 
            SQS.suc.cancel();
        time = 0;
    }
    
    /**
    * The head of the wait list (a circular two-way list).
    */
    private final static Activity waitList;

    static {
        waitList = new Activity(true) {
            public boolean condition() { return false; }
            public void startActions() {}
            public double duration() { return 0; }
            public void finishActions() {} 
        }; 
        waitList.pred = waitList.suc = waitList; 
    }
    
    /**
    * The head of the event list (a circular two-way list).
    */ 
    private final static Activity SQS;

    static {
        SQS = new Activity(true) {
            public boolean condition() { return false; }
            public void startActions() {}
            public double duration() { return 0; }
            public void finishActions() {} 
        };
        SQS.pred = SQS.suc =SQS;
     }
    
    /**
    * The current simulation time.
    */
    private static double time = 0;
    
    /**
    * The event time of this activity (is finish time).
    */
    private double eventTime;
    
    /**
    * The predecessor of this activity (in waitList or SQS).
    */
    private Activity pred;
    
    /**
    * The successor of this activity (in waitList or SQS).
    */
    private Activity suc;
}
