/*
  File: Event.java

  Originally written by Keld Helsgaun and released into the public domain.
  This may be used for any purposes whatsoever without acknowledgment.

  History:
  Date       Who                What
  5May2000   kh         Created public version
*/

package simulation.events;

/**
* The base class from which the classes 
* <tt>TimeEvent</tt> and <tt>StateEvent</tt> are inherited.
* <p>
* Class <tt>TimeEvent</tt> is used to describe the
* <b>time events</b> of a model, i.e., those events scheduled 
* to occur at specified points in time.
* <p>
* Class <tt>StateEvent</tt> is used to describe the
* <b>state events</b> of a model, i.e., those events scheduled 
* to occur to occur when the state of the system fulfills 
* a specified condition.
* <p>
* The <tt>actions</tt> method is used to describe the 
* actions associated with a class of events.
* When an event occurs, its action are executed.
* 
* @see simulation.events.TimeEvent
* @see simulation.events.StateEvent
* @see simulation.events.Simulation
*/

public abstract class Event {
    /**
    * The actions of this event.
    */
    protected abstract void actions();
	
    /**
    * Returns the current simulation time.
    */	
    public final static double time() { 
        return time; 
    }
	
	  /**
    * Runs a simulation for a specified period of simulated time.
    * <p> 
    * Time will start at 0 and jump from event time to event 
    * time until either this period is over, there are no 
    * more scheduled events, or the <tt>stopSimulation</tt> method 
    * is called.
    * <p>
    * All scheduled state events are held in list, and all 
    * scheduled time events are held in a list ordered by 
    * their associated event times.
    * <p> 
    * At each event time, the wait list is examined to see 
    * whether any state event has its condition fulfilled. 
    * In that case, the event is removed from the list and its 
    * actions are executed.
    * <p> 
    * When no more state events occur, time advances to 
    * the next imminent time event, and the associated 
    * actions of this time event are executed.
    * <p>
    * This continues until the simulation ends
    *  
    * @param <tt>period</tt> The length of the simulation period.
    */
    public final static void runSimulation(double period) {
        while (true) {
            for (StateEvent a = (StateEvent) waitList.suc; 
                 a != waitList; 
                 a = (StateEvent) a.suc) { 
                if (a.condition()) {  
                    a.cancel();
                    a.actions();
                    a = waitList;
                }
            }
            if (SQS.suc == SQS)
                break;
            TimeEvent ev = (TimeEvent) SQS.suc;
            time = ev.eventTime;
            ev.cancel();
            if (time > period)
                break;
            ev.actions();
        }
        stopSimulation();	
    }
    
    /**
    * Stops a simulation.
    */
	
    public final static void stopSimulation() {
        while (SQS.suc != SQS)
            SQS.suc.cancel();
        while (waitList.suc != waitList) 
            waitList.suc.cancel();
        time = 0;
    }
    
    /**
    * Cancels this scheduled event.
    * <p>
    * The activity is removed from the the list
    * of which it is a member (the event list or the wait list).
    * <p>
    * If it is not scheduled, the call has no effect.
    */
    public final void cancel() {
        if (suc != null) {
            suc.pred = pred;
            pred.suc = suc;
            suc = pred = null;
        }
    }
    
    /**
    * The head of the wait list (a circular two-way list).
    */ 
    static final TimeEvent SQS;

    static {
        SQS  = new TimeEvent() {
            protected void actions() {}
        };
        SQS.pred = SQS.suc = SQS; 
    }

    /**
    * The head of the event list (a circular two-way list).
    */ 
    static final StateEvent waitList;

    static {
        waitList = new StateEvent() {
            protected boolean condition() { return false; }
            protected void actions() {}
        };
        waitList.pred = waitList.suc = waitList; 
    }

    /**
    * The current simulation time.
    */
    static double time = 0;
    
    /**
    * The predecessor of this event (in waitList or SQS).
    */
    Event pred;
    
    /**
    * The successor of this activity (in waitList or SQS).
    */
    Event suc;
}
